// remove underscore from cache (because we would like to extend new object and not the reference from cache)
try {
    var rememberedCache = require.cache[require.resolve('underscore')];
    delete require.cache[require.resolve('underscore')]
} catch (e) {}

var _ = require('underscore');

// restore original cache state
try {
    require.cache[require.resolve('underscore')] = rememberedCache;
} catch (e) {}

var lookup = function (obj, key, defaultValue, defaultFalseValue) {
    var type = typeof key;
    if (type == 'string' || type == "number") {
        key = ("" + key).replace(/\[(.*?)\]/, function (m, key) { //handle case where [1] may occur
            return '.' + key.replace(/["']/g, ""); //strip quotes
        }).split('.');
    }
    for (var i = 0, l = key.length; i < l; i++) {
        //noinspection JSUnresolvedFunction
        if (_.has(obj, key[i])) {
            obj = obj[key[i]];
        } else {
            return defaultValue;
        }
    }
    if(defaultFalseValue !== undefined) {
        return (obj ? obj : defaultFalseValue);
    }
    return obj;
};

var lookupLength = function (obj, key, defaultValue) {
    var o = lookup(obj, key);

    if(o && (typeof o === 'array' || typeof o === 'object')) {
        return o.length;
    }
    return defaultValue;
};

/*  Copyright (C) 2012-2014  Kurt Milam - http://xioup.com | Source: https://gist.github.com/1868955
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **/
// Based conceptually on the _.extend() function in underscore.js ( see http://documentcloud.github.com/underscore/#extend for more details )
var deepExtend = function(obj) {
    var parentRE = /#{\s*?_\s*?}/,
        slice = Array.prototype.slice;

    _.each(slice.call(arguments, 1), function(source) {
        for (var prop in source) {
            if (_.isUndefined(obj[prop]) || _.isFunction(obj[prop]) || _.isNull(source[prop]) || _.isDate(source[prop])) {
                obj[prop] = source[prop];
            }
            else if (_.isString(source[prop]) && parentRE.test(source[prop])) {
                if (_.isString(obj[prop])) {
                    obj[prop] = source[prop].replace(parentRE, obj[prop]);
                }
            }
            else if (_.isArray(obj[prop]) || _.isArray(source[prop])){
                if (!_.isArray(obj[prop]) || !_.isArray(source[prop])){
                    throw new Error('Trying to combine an array with a non-array (' + prop + ')');
                } else {
                    obj[prop] = _.reject(_.deepExtend(_.clone(obj[prop]), source[prop]), function (item) { return _.isNull(item);});
                }
            }
            else if (_.isObject(obj[prop]) || _.isObject(source[prop])){
                if (!_.isObject(obj[prop]) || !_.isObject(source[prop])){
                    throw new Error('Trying to combine an object with a non-object (' + prop + ')');
                } else {
                    obj[prop] = _.deepExtend(_.clone(obj[prop]), source[prop]);
                }
            } else {
                obj[prop] = source[prop];
            }
        }
    });
    return obj;
};

//noinspection JSUnusedGlobalSymbols
_.mixin({
    lookup:       lookup,
    lookupLength: lookupLength,
    deepExtend:   deepExtend
});

/** @typedef {function} _.lookup */
/** @typedef {function} _.lookupLength */
/** @typedef {function} _.deepExtend */

module.exports = _;